@@ -1,12 +1,12 @@
 ;;; vertico-quick.el --- Quick keys for Vertico -*- lexical-binding: t -*-
 
-;; Copyright (C) 2021, 2022  Free Software Foundation, Inc.
+;; Copyright (C) 2021-2023 Free Software Foundation, Inc.
 
 ;; Author: Daniel Mendler <mail@daniel-mendler.de>
 ;; Maintainer: Daniel Mendler <mail@daniel-mendler.de>
 ;; Created: 2021
 ;; Version: 0.1
-;; Package-Requires: ((emacs "27.1") (vertico "0.29"))
+;; Package-Requires: ((emacs "27.1") (vertico "1.4"))
 ;; Homepage: https://github.com/minad/vertico
 
 ;; This file is part of GNU Emacs.
@@ -27,12 +27,12 @@
 ;;; Commentary:
 
 ;; This package is a Vertico extension, which prefixes candidates with
-;; quick keys. Typing these quick keys allows you to select the
-;; candidate in front of them. This is designed to be a faster
+;; quick keys.  Typing these quick keys allows you to select the
+;; candidate in front of them.  This is designed to be a faster
 ;; alternative to selecting a candidate with `vertico-next' and
 ;; `vertico-previous'.
-;; (define-key vertico-map "\M-q" #'vertico-quick-insert)
-;; (define-key vertico-map "\C-q" #'vertico-quick-exit)
+;; (keymap-set vertico-map "M-q" #'vertico-quick-insert)
+;; (keymap-set vertico-map "C-q" #'vertico-quick-exit)
 
 ;;; Code:
 
@@ -109,7 +109,7 @@ TWO is non-nil if two keys should be displayed."
                 (pcase-let ((`(,keys . ,events) (vertico-quick--keys first index start)))
                   (setq list (nconc events list))
                   (if (bound-and-true-p vertico-flat-mode)
-                      (setq keys (replace-regexp-in-string " " "" keys)
+                      (setq keys (string-replace " " "" keys)
                             cand (string-trim cand)
                             cand (substring cand (min (length cand) (length keys))))
                     (setq keys (concat keys (make-string (max 1 (- (length prefix) 2)) ?\s))))
