@@ -1,12 +1,12 @@
 ;;; vertico-reverse.el --- Reverse the Vertico display -*- lexical-binding: t -*-
 
-;; Copyright (C) 2021, 2022  Free Software Foundation, Inc.
+;; Copyright (C) 2021-2023 Free Software Foundation, Inc.
 
 ;; Author: Daniel Mendler <mail@daniel-mendler.de>
 ;; Maintainer: Daniel Mendler <mail@daniel-mendler.de>
 ;; Created: 2021
 ;; Version: 0.1
-;; Package-Requires: ((emacs "27.1") (vertico "0.29"))
+;; Package-Requires: ((emacs "27.1") (vertico "1.4"))
 ;; Homepage: https://github.com/minad/vertico
 
 ;; This file is part of GNU Emacs.
@@ -26,36 +26,45 @@
 
 ;;; Commentary:
 
-;; This package is a Vertico extension, which reverses the list of candidates.
+;; This package is a Vertico extension, which reverses the list of
+;; candidates.
 ;;
-;; The mode can be enabled globally or via `vertico-multiform-mode' per
-;; command or completion category. Alternatively the reverse display can be
-;; toggled temporarily if `vertico-multiform-mode' is enabled:
-;;
-;; (define-key vertico-map "\M-R" #'vertico-multiform-reverse)
+;; The mode can be enabled globally or via `vertico-multiform-mode'
+;; per command or completion category.  Alternatively the reverse
+;; display can be toggled temporarily with M-R if
+;; `vertico-multiform-mode' is enabled.
 
 ;;; Code:
 
 (require 'vertico)
 
-(defvar vertico-reverse-map
-  (let ((map (make-sparse-keymap)))
-    (define-key map [remap beginning-of-buffer] #'vertico-last)
-    (define-key map [remap minibuffer-beginning-of-buffer] #'vertico-last)
-    (define-key map [remap end-of-buffer] #'vertico-first)
-    (define-key map [remap scroll-down-command] #'vertico-scroll-up)
-    (define-key map [remap scroll-up-command] #'vertico-scroll-down)
-    (define-key map [remap next-line] #'vertico-previous)
-    (define-key map [remap previous-line] #'vertico-next)
-    (define-key map [remap next-line-or-history-element] #'vertico-previous)
-    (define-key map [remap previous-line-or-history-element] #'vertico-next)
-    (define-key map [remap backward-paragraph] #'vertico-next-group)
-    (define-key map [remap forward-paragraph] #'vertico-previous-group)
-    map)
-  "Additional keymap activated in reverse mode.")
+(defvar-keymap vertico-reverse-map
+  :doc "Additional keymap activated in reverse mode."
+  "<remap> <beginning-of-buffer>" #'vertico-last
+  "<remap> <minibuffer-beginning-of-buffer>" #'vertico-last
+  "<remap> <end-of-buffer>" #'vertico-first
+  "<remap> <scroll-down-command>" #'vertico-scroll-up
+  "<remap> <scroll-up-command>" #'vertico-scroll-down
+  "<remap> <next-line>" #'vertico-previous
+  "<remap> <previous-line>" #'vertico-next
+  "<remap> <next-line-or-history-element>" #'vertico-previous
+  "<remap> <previous-line-or-history-element>" #'vertico-next
+  "<remap> <backward-paragraph>" #'vertico-next-group
+  "<remap> <forward-paragraph>" #'vertico-previous-group)
 
-(defun vertico-reverse--display-candidates (lines)
-  "Display LINES in reverse."
+;;;###autoload
+(define-minor-mode vertico-reverse-mode
+  "Reverse the Vertico display."
+  :global t :group 'vertico
+  ;; Reset overlays
+  (dolist (buf (buffer-list))
+    (when-let (ov (buffer-local-value 'vertico--candidates-ov buf))
+      (overlay-put ov 'before-string nil)))
+  (if vertico-reverse-mode
+      (add-to-list 'minor-mode-map-alist `(vertico--input . ,vertico-reverse-map))
+    (setq minor-mode-map-alist (delete `(vertico--input . ,vertico-reverse-map) minor-mode-map-alist))))
+
+(cl-defmethod vertico--display-candidates (lines &context (vertico-reverse-mode (eql t)))
   (move-overlay vertico--candidates-ov (point-min) (point-min))
   (setq lines (nreverse lines))
   (unless (eq vertico-resize t)
@@ -66,21 +75,5 @@
     (overlay-put vertico--candidates-ov 'after-string nil))
   (vertico--resize-window (length lines)))
 
-;;;###autoload
-(define-minor-mode vertico-reverse-mode
-  "Reverse the Vertico display."
-  :global t :group 'vertico
-  ;; Reset overlays
-  (dolist (buf (buffer-list))
-    (when-let (ov (buffer-local-value 'vertico--candidates-ov buf))
-      (overlay-put ov 'before-string nil)))
-  (cond
-   (vertico-reverse-mode
-    (add-to-list 'minor-mode-map-alist `(vertico--input . ,vertico-reverse-map))
-    (advice-add #'vertico--display-candidates :override #'vertico-reverse--display-candidates))
-   (t
-    (setq minor-mode-map-alist (delete `(vertico--input . ,vertico-reverse-map) minor-mode-map-alist))
-    (advice-remove #'vertico--display-candidates #'vertico-reverse--display-candidates))))
-
 (provide 'vertico-reverse)
 ;;; vertico-reverse.el ends here
