@@ -1,13 +1,14 @@
 ;;; vertico.el --- VERTical Interactive COmpletion -*- lexical-binding: t -*-
 
-;; Copyright (C) 2021, 2022  Free Software Foundation, Inc.
+;; Copyright (C) 2021-2023 Free Software Foundation, Inc.
 
 ;; Author: Daniel Mendler <mail@daniel-mendler.de>
 ;; Maintainer: Daniel Mendler <mail@daniel-mendler.de>
 ;; Created: 2021
-;; Version: 0.29
-;; Package-Requires: ((emacs "27.1"))
+;; Version: 1.4
+;; Package-Requires: ((emacs "27.1") (compat "29.1.4.0"))
 ;; Homepage: https://github.com/minad/vertico
+;; Keywords: convenience, files, matching, completion
 
 ;; This file is part of GNU Emacs.
 
@@ -27,19 +28,22 @@
 ;;; Commentary:
 
 ;; Vertico provides a performant and minimalistic vertical completion UI
-;; based on the default completion system. By reusing the built-in
+;; based on the default completion system.  By reusing the built-in
 ;; facilities, Vertico achieves full compatibility with built-in Emacs
 ;; completion commands and completion tables.
 
 ;;; Code:
 
-(require 'seq)
+(require 'compat)
 (eval-when-compile
   (require 'cl-lib)
   (require 'subr-x))
 
 (defgroup vertico nil
   "VERTical Interactive COmpletion."
+  :link '(info-link :tag "Info Manual" "(vertico)")
+  :link '(url-link :tag "Homepage" "https://github.com/minad/vertico")
+  :link '(emacs-library-link :tag "Library Source" "vertico.el")
   :group 'convenience
   :group 'minibuffer
   :prefix "vertico-")
@@ -57,12 +61,19 @@
 
 (defcustom vertico-count 10
   "Maximal number of candidates to show."
-  :type 'integer)
+  :type 'natnum)
+
+(defcustom vertico-preselect 'directory
+  "Configure if the prompt or first candidate is preselected.
+- prompt: Always select the prompt.
+- first: Always select the first candidate.
+- directory: Like first, but select the prompt if it is a directory."
+  :type '(choice (const prompt) (const first) (const directory)))
 
 (defcustom vertico-scroll-margin 2
   "Number of lines at the top and bottom when scrolling.
 The value should lie between 0 and vertico-count/2."
-  :type 'integer)
+  :type 'natnum)
 
 (defcustom vertico-resize resize-mini-windows
   "How to resize the Vertico minibuffer window, see `resize-mini-windows'."
@@ -75,7 +86,7 @@ The value should lie between 0 and vertico-count/2."
   :type 'boolean)
 
 (defcustom vertico-multiline
-  (cons #("⤶" 0 1 (face vertico-multiline)) #("…" 0 1 (face vertico-multiline)))
+  (cons #("↲" 0 1 (face vertico-multiline)) #("…" 0 1 (face vertico-multiline)))
   "Replacements for multiline strings."
   :type '(cons (string :tag "Newline") (string :tag "Truncation")))
 
@@ -110,25 +121,24 @@ The value should lie between 0 and vertico-count/2."
 (defface vertico-current '((t :inherit highlight :extend t))
   "Face used to highlight the currently selected candidate.")
 
-(defvar vertico-map
-  (let ((map (make-composed-keymap nil minibuffer-local-map)))
-    (define-key map [remap beginning-of-buffer] #'vertico-first)
-    (define-key map [remap minibuffer-beginning-of-buffer] #'vertico-first)
-    (define-key map [remap end-of-buffer] #'vertico-last)
-    (define-key map [remap scroll-down-command] #'vertico-scroll-down)
-    (define-key map [remap scroll-up-command] #'vertico-scroll-up)
-    (define-key map [remap next-line] #'vertico-next)
-    (define-key map [remap previous-line] #'vertico-previous)
-    (define-key map [remap next-line-or-history-element] #'vertico-next)
-    (define-key map [remap previous-line-or-history-element] #'vertico-previous)
-    (define-key map [remap backward-paragraph] #'vertico-previous-group)
-    (define-key map [remap forward-paragraph] #'vertico-next-group)
-    (define-key map [remap exit-minibuffer] #'vertico-exit)
-    (define-key map [remap kill-ring-save] #'vertico-save)
-    (define-key map "\M-\r" #'vertico-exit-input)
-    (define-key map "\t" #'vertico-insert)
-    map)
-  "Vertico minibuffer keymap derived from `minibuffer-local-map'.")
+(defvar-keymap vertico-map
+  :doc "Vertico minibuffer keymap derived from `minibuffer-local-map'."
+  :parent minibuffer-local-map
+  "<remap> <beginning-of-buffer>" #'vertico-first
+  "<remap> <minibuffer-beginning-of-buffer>" #'vertico-first
+  "<remap> <end-of-buffer>" #'vertico-last
+  "<remap> <scroll-down-command>" #'vertico-scroll-down
+  "<remap> <scroll-up-command>" #'vertico-scroll-up
+  "<remap> <next-line>" #'vertico-next
+  "<remap> <previous-line>" #'vertico-previous
+  "<remap> <next-line-or-history-element>" #'vertico-next
+  "<remap> <previous-line-or-history-element>" #'vertico-previous
+  "<remap> <backward-paragraph>" #'vertico-previous-group
+  "<remap> <forward-paragraph>" #'vertico-next-group
+  "<remap> <exit-minibuffer>" #'vertico-exit
+  "<remap> <kill-ring-save>" #'vertico-save
+  "M-RET" #'vertico-exit-input
+  "TAB" #'vertico-insert)
 
 (defvar-local vertico--highlight #'identity
   "Deferred candidate highlighting function.")
@@ -182,19 +192,24 @@ The value should lie between 0 and vertico-count/2."
   "Recompute history hash table and return it."
   (or (and (equal (car vertico--history-hash) vertico--base) (cdr vertico--history-hash))
       (let* ((base vertico--base)
-             (base-size (length base))
+             (base-len (length base))
              (hist (and (not (eq minibuffer-history-variable t)) ;; Disabled for `t'.
                         (symbol-value minibuffer-history-variable)))
-             (hash (make-hash-table :test #'equal :size (length hist))))
+             (hash (make-hash-table :test #'equal :size (length hist)))
+             (file-p (and (> base-len 0) ;; Step-wise completion, unlike `project-find-file'
+                          (eq minibuffer-history-variable 'file-name-history)))
+             (curr-file (when-let ((win (and file-p (minibuffer-selected-window)))
+                                   (file (buffer-file-name (window-buffer win))))
+                          (abbreviate-file-name file))))
         (cl-loop for elem in hist for index from 0 do
-                 (when (or (= base-size 0)
-                           (and (>= (length elem) base-size)
-                                (eq t (compare-strings base 0 base-size elem 0 base-size))))
-                   (let ((file-sep (and (eq minibuffer-history-variable 'file-name-history)
-                                        (string-match-p "/" elem base-size))))
+                 (when (and (not (equal curr-file elem)) ;; Deprioritize current file
+                            (or (= base-len 0)
+                                (and (>= (length elem) base-len)
+                                     (eq t (compare-strings base 0 base-len elem 0 base-len)))))
+                   (let ((file-sep (and file-p (string-search "/" elem base-len))))
                      ;; Drop base string from history elements & special file handling.
-                     (when (or (> base-size 0) file-sep)
-                       (setq elem (substring elem base-size (and file-sep (1+ file-sep)))))
+                     (when (or (> base-len 0) file-sep)
+                       (setq elem (substring elem base-len (and file-sep (1+ file-sep)))))
                      (unless (gethash elem hash) (puthash elem index hash)))))
         (cdr (setq vertico--history-hash (cons base hash))))))
 
@@ -216,15 +231,11 @@ The function is configured by BY, BSIZE, BINDEX, BPRED and PRED."
      (let* ((buckets (make-vector ,bsize nil))
             ,@(and (eq (car by) 'history) '((hhash (vertico--history-hash)) (hcands))))
        (dolist (% candidates)
-         ,(if (eq (car by) 'history)
-              ;; Find recent candidates or fill buckets
-              `(if-let (idx (gethash % hhash))
-                   (push (cons idx %) hcands)
-                 (let ((idx (min ,(1- bsize) ,bindex)))
-                   (aset buckets idx (cons % (aref buckets idx)))))
-            ;; Fill buckets
-            `(let ((idx (min ,(1- bsize) ,bindex)))
-               (aset buckets idx (cons % (aref buckets idx))))))
+         ;; Find recent candidate in history or fill bucket
+         (,@(if (not (eq (car by) 'history)) `(progn)
+              `(if-let (idx (gethash % hhash)) (push (cons idx %) hcands)))
+          (let ((idx (min ,(1- bsize) ,bindex)))
+            (aset buckets idx (cons % (aref buckets idx))))))
        (nconc ,@(and (eq (car by) 'history) '((vertico--sort-decorated hcands)))
               (mapcan (lambda (bucket) (sort bucket #',bpred))
                       (nbutlast (append buckets nil)))
@@ -245,8 +256,8 @@ The function is configured by BY, BSIZE, BINDEX, BPRED and PRED."
                      (plist-get completion-extra-properties :annotation-function)))
         (cl-loop for cand in cands collect
                  (let ((suffix (or (funcall ann cand) "")))
-                   ;; The default completion UI adds the `completions-annotations' face
-                   ;; if no other faces are present.
+                   ;; The default completion UI adds the `completions-annotations'
+                   ;; face if no other faces are present.
                    (unless (text-property-not-all 0 (length suffix) 'face nil suffix)
                      (setq suffix (propertize suffix 'face 'completions-annotations)))
                    (list cand "" suffix)))
@@ -255,13 +266,12 @@ The function is configured by BY, BSIZE, BINDEX, BPRED and PRED."
 (defun vertico--move-to-front (elem list)
   "Move ELEM to front of LIST."
   (if-let (found (member elem list))
-      (let ((head (list (car found))))
-        (nconc head (delq (setcar found nil) list)))
+      (nconc (list (car found)) (delq (setcar found nil) list))
     list))
 
 ;; bug#47711: Deferred highlighting for `completion-all-completions'
-;; XXX There is one complication: `completion--twq-all' already adds `completions-common-part'.
-;; See below `vertico--candidate'.
+;; XXX There is one complication: `completion--twq-all' already adds
+;; `completions-common-part'.  See below `vertico--candidate'.
 (defun vertico--all-completions (&rest args)
   "Compute all completions for ARGS with deferred highlighting."
   (cl-letf* ((orig-pcm (symbol-function #'completion-pcm--hilit-commonality))
@@ -307,31 +317,22 @@ The function is configured by BY, BSIZE, BINDEX, BPRED and PRED."
       (vertico--metadata-get 'display-sort-function)
       vertico-sort-function))
 
-(defun vertico--filter-files (files)
-  "Filter FILES by `completion-ignored-extensions'."
-  (let ((re (concat "\\(?:\\(?:\\`\\|/\\)\\.\\.?/\\|"
-                    (regexp-opt completion-ignored-extensions)
-                    "\\)\\'")))
-    (or (seq-remove (lambda (x) (string-match-p re x)) files) files)))
-
 (defun vertico--recompute (pt content)
   "Recompute state given PT and CONTENT."
-  (pcase-let* ((before (substring content 0 pt))
+  (pcase-let* ((table minibuffer-completion-table)
+               (pred minibuffer-completion-predicate)
+               (before (substring content 0 pt))
                (after (substring content pt))
-               ;; bug#47678: `completion-boundaries` fails for `partial-completion`
+               ;; bug#47678: `completion-boundaries' fails for `partial-completion'
                ;; if the cursor is moved between the slashes of "~//".
                ;; See also marginalia.el which has the same issue.
                (bounds (or (condition-case nil
-                               (completion-boundaries
-                                before minibuffer-completion-table
-                                minibuffer-completion-predicate after)
+                               (completion-boundaries before table pred after)
                              (t (cons 0 (length after))))))
                (field (substring content (car bounds) (+ pt (cdr bounds))))
                ;; `minibuffer-completing-file-name' has been obsoleted by the completion category
                (completing-file (eq 'file (vertico--metadata-get 'category)))
-               (`(,all . ,hl) (vertico--all-completions
-                               content minibuffer-completion-table
-                               minibuffer-completion-predicate pt vertico--metadata))
+               (`(,all . ,hl) (vertico--all-completions content table pred pt vertico--metadata))
                (base (or (when-let (z (last all)) (prog1 (cdr z) (setcdr z nil))) 0))
                (vertico--base (substring content 0 base))
                (def (or (car-safe minibuffer-default) minibuffer-default))
@@ -339,7 +340,7 @@ The function is configured by BY, BSIZE, BINDEX, BPRED and PRED."
     ;; Filter the ignored file extensions. We cannot use modified predicate for this filtering,
     ;; since this breaks the special casing in the `completion-file-name-table' for `file-exists-p'
     ;; and `file-directory-p'.
-    (when completing-file (setq all (vertico--filter-files all)))
+    (when completing-file (setq all (completion-pcm--filename-try-filter all)))
     ;; Sort using the `display-sort-function' or the Vertico sort functions
     (setq all (delete-consecutive-dups (funcall (or (vertico--sort-function) #'identity) all)))
     ;; Move special candidates: "field" appears at the top, before "field/", before default value
@@ -359,22 +360,17 @@ The function is configured by BY, BSIZE, BINDEX, BPRED and PRED."
       (vertico--candidates . ,all)
       (vertico--total . ,(length all))
       (vertico--highlight . ,hl)
-      (vertico--allow-prompt . ,(or def-missing
+      (vertico--allow-prompt . ,(or def-missing (eq vertico-preselect 'prompt)
                                     (memq minibuffer--require-match
                                           '(nil confirm confirm-after-completion))))
       (vertico--lock-candidate . ,lock)
       (vertico--groups . ,(cadr groups))
       (vertico--all-groups . ,(or (caddr groups) vertico--all-groups))
-      ;; Compute new index. Select the prompt under these conditions:
-      ;; * If there are no candidates
-      ;; * If the default is missing from the candidate list.
-      ;; * For matching content, as long as the full content
-      ;;   after the boundary is empty, including content after point.
       (vertico--index . ,(or lock
-                             (if (or def-missing (not all)
-                                     (and (= (length vertico--base) (length content))
-                                          (test-completion content minibuffer-completion-table
-                                                           minibuffer-completion-predicate)))
+                             (if (or def-missing (eq vertico-preselect 'prompt) (not all)
+                                     (and completing-file (eq vertico-preselect 'directory)
+                                          (= (length vertico--base) (length content))
+                                          (test-completion content table pred)))
                                  -1 0))))))
 
 (defun vertico--cycle (list n)
@@ -420,21 +416,17 @@ The function is configured by BY, BSIZE, BINDEX, BPRED and PRED."
   "Return t if PATH is a remote path."
   (string-match-p "\\`/[^/|:]+:" (substitute-in-file-name path)))
 
-(defun vertico--prepare ()
-  "Ensure that the state is prepared before running the next command."
-  (when (and (symbolp this-command) (string-prefix-p "vertico-" (symbol-name this-command)))
-    (vertico--update)))
-
 (defun vertico--update (&optional interruptible)
   "Update state, optionally INTERRUPTIBLE."
   (let* ((pt (max 0 (- (point) (minibuffer-prompt-end))))
          (content (minibuffer-contents-no-properties))
          (input (cons content pt)))
     (unless (or (and interruptible (input-pending-p)) (equal vertico--input input))
-      ;; Redisplay the minibuffer such that the input becomes immediately
-      ;; visible before the expensive candidate recomputation (Issue #89).
-      ;; Do not redisplay during initialization, since this leads to flicker.
-      (when (and interruptible (consp vertico--input)) (redisplay))
+      ;; Redisplay to make input immediately visible before expensive candidate
+      ;; recomputation (#89).  No redisplay during init because of flicker.
+      (when (and interruptible (consp vertico--input))
+        ;; Prevent recursive exhibit from timer (`consult-vertico--refresh').
+        (cl-letf (((symbol-function #'vertico--exhibit) #'ignore)) (redisplay)))
       (pcase (let ((vertico--metadata (completion-metadata (substring content 0 pt)
                                                            minibuffer-completion-table
                                                            minibuffer-completion-predicate)))
@@ -456,14 +448,17 @@ The function is configured by BY, BSIZE, BINDEX, BPRED and PRED."
   (let ((end (length str)) (pos 0) chunks)
     (while (< pos end)
       (let ((nextd (next-single-property-change pos 'display str end))
-            (display (get-text-property pos 'display str)))
-        (if (stringp display)
-            (progn (push display chunks) (setq pos nextd))
+            (disp (get-text-property pos 'display str)))
+        (if (stringp disp)
+            (let ((face (get-text-property pos 'face str)))
+              (when face
+                (add-face-text-property 0 (length disp) face t (setq disp (concat disp))))
+              (setq pos nextd chunks (cons disp chunks)))
           (while (< pos nextd)
             (let ((nexti (next-single-property-change pos 'invisible str nextd)))
-              (unless (get-text-property pos 'invisible str)
-                (unless (and (= pos 0) (= nexti end)) ;; full string -> avoid allocation
-                  (push (substring str pos nexti) chunks)))
+              (unless (or (get-text-property pos 'invisible str)
+                          (and (= pos 0) (= nexti end))) ;; full string -> no allocation
+                  (push (substring str pos nexti) chunks))
               (setq pos nexti))))))
     (if chunks (apply #'concat (nreverse chunks)) str)))
 
@@ -480,13 +475,6 @@ The function is configured by BY, BSIZE, BINDEX, BPRED and PRED."
      (replace-regexp-in-string "\\`[\t\n ]+\\|[\t\n ]+\\'" ""))
    max-width 0 nil (cdr vertico-multiline)))
 
-(defun vertico--format-candidate (cand prefix suffix index _start)
-  "Format CAND given PREFIX, SUFFIX and INDEX."
-  (setq cand (vertico--display-string (concat prefix cand suffix "\n")))
-  (when (= index vertico--index)
-    (add-face-text-property 0 (length cand) 'vertico-current 'append cand))
-  cand)
-
 (defun vertico--compute-scroll ()
   "Compute new scroll position."
   (let ((off (max (min vertico-scroll-margin (/ vertico-count 2)) 0))
@@ -505,68 +493,6 @@ The function is configured by BY, BSIZE, BINDEX, BPRED and PRED."
     (vertico--remove-face 0 (length title) 'completions-first-difference title))
   (format (concat vertico-group-format "\n") title))
 
-(defun vertico--arrange-candidates ()
-  "Arrange candidates."
-  (vertico--compute-scroll)
-  (let ((curr-line 0) lines)
-    ;; Compute group titles
-    (let* (title (index vertico--scroll)
-           (group-fun (and vertico-group-format (vertico--metadata-get 'group-function)))
-           (candidates
-            (thread-last (seq-subseq vertico--candidates index
-                                     (min (+ index vertico-count) vertico--total))
-              (funcall vertico--highlight)
-              (vertico--affixate))))
-      (pcase-dolist ((and cand `(,str . ,_)) candidates)
-        (when-let (new-title (and group-fun (funcall group-fun str nil)))
-          (unless (equal title new-title)
-            (setq title new-title)
-            (push (vertico--format-group-title title str) lines))
-          (setcar cand (funcall group-fun str 'transform)))
-        (when (= index vertico--index)
-          (setq curr-line (length lines)))
-        (push (cons index cand) lines)
-        (setq index (1+ index))))
-    ;; Drop excess lines
-    (setq lines (nreverse lines))
-    (cl-loop for count from (length lines) above vertico-count do
-             (if (< curr-line (/ count 2))
-                 (nbutlast lines)
-               (setq curr-line (1- curr-line) lines (cdr lines))))
-    ;; Format candidates
-    (let ((max-width (- (vertico--window-width) 4)) start)
-      (cl-loop for line on lines do
-               (pcase (car line)
-                 (`(,index ,cand ,prefix ,suffix)
-                  (setq start (or start index))
-                  (when (string-match-p "\n" cand)
-                    (setq cand (vertico--truncate-multiline cand max-width)))
-                  (setcar line (vertico--format-candidate cand prefix suffix index start))))))
-    lines))
-
-(defun vertico--display-candidates (lines)
-  "Update candidates overlay `vertico--candidates-ov' with LINES."
-  (move-overlay vertico--candidates-ov (point-max) (point-max))
-  (overlay-put vertico--candidates-ov 'after-string
-               (apply #'concat #(" " 0 1 (cursor t)) (and lines "\n") lines))
-  (vertico--resize-window (length lines)))
-
-(defun vertico--resize-window (height)
-  "Resize active minibuffer window to HEIGHT."
-  (setq-local truncate-lines (< (point) (* 0.8 (vertico--window-width)))
-              resize-mini-windows 'grow-only
-              max-mini-window-height 1.0)
-  (unless (frame-root-window-p (active-minibuffer-window))
-    (unless vertico-resize
-      (setq height (max height vertico-count)))
-    (let* ((window-resize-pixelwise t)
-           (dp (- (max (cdr (window-text-pixel-size))
-                       (* (default-line-height) (1+ height)))
-                  (window-pixel-height))))
-      (when (or (and (> dp 0) (/= height 0))
-                (and (< dp 0) (eq vertico-resize t)))
-        (window-resize nil dp nil nil 'pixelwise)))))
-
 (defun vertico--format-count ()
   "Format the count string."
   (format (car vertico-count-format)
@@ -594,7 +520,7 @@ The function is configured by BY, BSIZE, BINDEX, BPRED and PRED."
   (while (< beg end)
     (let ((next (next-single-property-change beg 'face obj end)))
       (when-let (val (get-text-property beg 'face obj))
-        (put-text-property beg next 'face (remq face (if (listp val) val (list val))) obj))
+        (put-text-property beg next 'face (remq face (ensure-list val)) obj))
       (setq beg next))))
 
 (defun vertico--exhibit ()
@@ -612,6 +538,120 @@ The function is configured by BY, BSIZE, BINDEX, BPRED and PRED."
              (min index (1- vertico--total)))
         vertico--lock-candidate (or (>= vertico--index 0) vertico--allow-prompt)))
 
+(defun vertico--candidate (&optional hl)
+  "Return current candidate string with optional highlighting if HL is non-nil."
+  (let ((content (substring (or (car-safe vertico--input) (minibuffer-contents-no-properties)))))
+    (cond
+     ((>= vertico--index 0)
+      (let ((cand (substring (nth vertico--index vertico--candidates))))
+        ;; XXX Drop the completions-common-part face which is added by the
+        ;; `completion--twq-all' hack.  This should better be fixed in Emacs
+        ;; itself, the corresponding code is already marked with a FIXME.
+        (vertico--remove-face 0 (length cand) 'completions-common-part cand)
+        (concat vertico--base
+                (if hl (car (funcall vertico--highlight (list cand))) cand))))
+     ((and (equal content "") (or (car-safe minibuffer-default) minibuffer-default)))
+     (t content))))
+
+(defun vertico--match-p (input)
+  "Return t if INPUT is a valid match."
+  (let ((rm minibuffer--require-match))
+    (or (memq rm '(nil confirm-after-completion))
+        (equal "" input) ;; Null completion, returns default value
+        (and (functionp rm) (funcall rm input)) ;; Emacs 29 supports functions
+        (test-completion input minibuffer-completion-table minibuffer-completion-predicate)
+        (if (eq rm 'confirm) (eq (ignore-errors (read-char "Confirm")) 13)
+          (minibuffer-message "Match required") nil))))
+
+(cl-defgeneric vertico--format-candidate (cand prefix suffix index _start)
+  "Format CAND given PREFIX, SUFFIX and INDEX."
+  (setq cand (vertico--display-string (concat prefix cand suffix "\n")))
+  (when (= index vertico--index)
+    (add-face-text-property 0 (length cand) 'vertico-current 'append cand))
+  cand)
+
+(cl-defgeneric vertico--arrange-candidates ()
+  "Arrange candidates."
+  (vertico--compute-scroll)
+  (let ((curr-line 0) lines)
+    ;; Compute group titles
+    (let* (title (index vertico--scroll)
+           (group-fun (and vertico-group-format (vertico--metadata-get 'group-function)))
+           (candidates
+            (thread-last (seq-subseq vertico--candidates index
+                                     (min (+ index vertico-count) vertico--total))
+              (funcall vertico--highlight)
+              (vertico--affixate))))
+      (pcase-dolist ((and cand `(,str . ,_)) candidates)
+        (when-let (new-title (and group-fun (funcall group-fun str nil)))
+          (unless (equal title new-title)
+            (setq title new-title)
+            (push (vertico--format-group-title title str) lines))
+          (setcar cand (funcall group-fun str 'transform)))
+        (when (= index vertico--index)
+          (setq curr-line (length lines)))
+        (push (cons index cand) lines)
+        (cl-incf index)))
+    ;; Drop excess lines
+    (setq lines (nreverse lines))
+    (cl-loop for count from (length lines) above vertico-count do
+             (if (< curr-line (/ count 2))
+                 (nbutlast lines)
+               (setq curr-line (1- curr-line) lines (cdr lines))))
+    ;; Format candidates
+    (let ((max-width (- (vertico--window-width) 4)) start)
+      (cl-loop for line on lines do
+               (pcase (car line)
+                 (`(,index ,cand ,prefix ,suffix)
+                  (setq start (or start index))
+                  (when (string-search "\n" cand)
+                    (setq cand (vertico--truncate-multiline cand max-width)))
+                  (setcar line (vertico--format-candidate cand prefix suffix index start))))))
+    lines))
+
+(cl-defgeneric vertico--display-candidates (lines)
+  "Update candidates overlay `vertico--candidates-ov' with LINES."
+  (move-overlay vertico--candidates-ov (point-max) (point-max))
+  (overlay-put vertico--candidates-ov 'after-string
+               (apply #'concat #(" " 0 1 (cursor t)) (and lines "\n") lines))
+  (vertico--resize-window (length lines)))
+
+(cl-defgeneric vertico--resize-window (height)
+  "Resize active minibuffer window to HEIGHT."
+  (setq-local truncate-lines (< (point) (* 0.8 (vertico--window-width)))
+              resize-mini-windows 'grow-only
+              max-mini-window-height 1.0)
+  (unless (frame-root-window-p (active-minibuffer-window))
+    (unless vertico-resize (setq height (max height vertico-count)))
+    (let ((dp (- (max (cdr (window-text-pixel-size))
+                      (* (default-line-height) (1+ height)))
+                 (window-pixel-height))))
+      (when (or (and (> dp 0) (/= height 0))
+                (and (< dp 0) (eq vertico-resize t)))
+        (window-resize nil dp nil nil 'pixelwise)))))
+
+(cl-defgeneric vertico--prepare ()
+  "Ensure that the state is prepared before running the next command."
+  (when (and (symbolp this-command) (string-prefix-p "vertico-" (symbol-name this-command)))
+    (vertico--update)))
+
+(cl-defgeneric vertico--setup ()
+  "Setup completion UI."
+  (setq-local scroll-margin 0
+              vertico--input t
+              completion-auto-help nil
+              completion-show-inline-help nil
+              vertico--candidates-ov (make-overlay (point-max) (point-max) nil t t)
+              vertico--count-ov (make-overlay (point-min) (point-min) nil t t))
+  (overlay-put vertico--count-ov 'priority 1) ;; For `minibuffer-depth-indicate-mode'
+  (use-local-map vertico-map)
+  (add-hook 'pre-command-hook #'vertico--prepare nil 'local)
+  (add-hook 'post-command-hook #'vertico--exhibit nil 'local))
+
+(cl-defgeneric vertico--advice (&rest app)
+  "Advice for completion function, apply APP."
+  (minibuffer-with-setup-hook #'vertico--setup (apply app)))
+
 (defun vertico-first ()
   "Go to first candidate, or to the prompt when the first candidate is selected."
   (interactive)
@@ -648,15 +688,6 @@ The function is configured by BY, BSIZE, BINDEX, BPRED and PRED."
   (interactive "p")
   (vertico-next (- (or n 1))))
 
-(defun vertico--match-p (input)
-  "Return t if INPUT is a valid match."
-  (or (memq minibuffer--require-match '(nil confirm-after-completion))
-      (equal "" input) ;; Null completion, returns default value
-      (test-completion input minibuffer-completion-table minibuffer-completion-predicate)
-      (if (eq minibuffer--require-match 'confirm)
-          (eq (ignore-errors (read-char "Confirm")) 13)
-        (and (minibuffer-message "Match required") nil))))
-
 (defun vertico-exit (&optional arg)
   "Exit minibuffer with current candidate or input if prefix ARG is given."
   (interactive "P")
@@ -712,48 +743,18 @@ When the prefix argument is 0, the group order is reset."
     (let ((vertico--index (max 0 vertico--index)))
       (insert (prog1 (vertico--candidate) (delete-minibuffer-contents))))))
 
-(defun vertico--candidate (&optional hl)
-  "Return current candidate string with optional highlighting if HL is non-nil."
-  (let ((content (substring (or (car-safe vertico--input) (minibuffer-contents-no-properties)))))
-    (cond
-     ((>= vertico--index 0)
-      (let ((cand (substring (nth vertico--index vertico--candidates))))
-        ;; XXX Drop the completions-common-part face which is added by `completion--twq-all'.
-        ;; This is a hack in Emacs and should better be fixed in Emacs itself, the corresponding
-        ;; code is already marked with a FIXME. Should this be reported as a bug?
-        (vertico--remove-face 0 (length cand) 'completions-common-part cand)
-        (concat vertico--base
-                (if hl (car (funcall vertico--highlight (list cand))) cand))))
-     ((and (equal content "") (or (car-safe minibuffer-default) minibuffer-default)))
-     (t content))))
-
-(defun vertico--setup ()
-  "Setup completion UI."
-  (setq vertico--input t
-        vertico--candidates-ov (make-overlay (point-max) (point-max) nil t t)
-        vertico--count-ov (make-overlay (point-min) (point-min) nil t t))
-  ;; Set priority for compatibility with `minibuffer-depth-indicate-mode'
-  (overlay-put vertico--count-ov 'priority 1)
-  (setq-local completion-auto-help nil
-              completion-show-inline-help nil)
-  (use-local-map vertico-map)
-  (add-hook 'pre-command-hook #'vertico--prepare nil 'local)
-  (add-hook 'post-command-hook #'vertico--exhibit nil 'local))
-
-(defun vertico--advice (&rest args)
-  "Advice for completion function, receiving ARGS."
-  (minibuffer-with-setup-hook #'vertico--setup (apply args)))
-
 ;;;###autoload
 (define-minor-mode vertico-mode
   "VERTical Interactive COmpletion."
   :global t :group 'vertico
-  (if vertico-mode
-      (progn
-        (advice-add #'completing-read-default :around #'vertico--advice)
-        (advice-add #'completing-read-multiple :around #'vertico--advice))
-    (advice-remove #'completing-read-default #'vertico--advice)
-    (advice-remove #'completing-read-multiple #'vertico--advice)))
+  (dolist (fun '(completing-read-default completing-read-multiple))
+    (if vertico-mode
+        (advice-add fun :around #'vertico--advice)
+      (advice-remove fun #'vertico--advice))))
+
+(defun vertico--command-p (_sym buffer)
+  "Return non-nil if Vertico is active in BUFFER."
+  (buffer-local-value 'vertico--input buffer))
 
 ;; Emacs 28: Do not show Vertico commands in M-X
 (dolist (sym '(vertico-next vertico-next-group vertico-previous vertico-previous-group
@@ -761,9 +762,5 @@ When the prefix argument is 0, the group order is reset."
                vertico-exit-input vertico-save vertico-first vertico-last))
   (put sym 'completion-predicate #'vertico--command-p))
 
-(defun vertico--command-p (_sym buffer)
-  "Return non-nil if Vertico is active in BUFFER."
-  (buffer-local-value 'vertico--input buffer))
-
 (provide 'vertico)
 ;;; vertico.el ends here
